#ifndef PROG_FEATURES_H
#define PROG_FEATURES_H
struct prog_feature {
	char *prog_name;
	__u32 features;
};

static struct prog_feature prog_features[] = {
	{"xdpfilt_dny_udp.o", 0x42000000},
	{"xdpfilt_dny_tcp.o", 0x41000000},
	{"xdpfilt_dny_ip.o", 0x4c000000},
	{"xdpfilt_dny_eth.o", 0x50000000},
	{"xdpfilt_dny_all.o", 0x5f000000},
	{"xdpfilt_alw_udp.o", 0x22000000},
	{"xdpfilt_alw_tcp.o", 0x21000000},
	{"xdpfilt_alw_ip.o", 0x2c000000},
	{"xdpfilt_alw_eth.o", 0x30000000},
	{"xdpfilt_alw_all.o", 0x3f000000},
	{}
};
#endif
